import inspect

import pygame
import pygame_phyics.util as _util
from typing import List
from pygame_phyics.objects import *
from pygame_phyics.manger import Manger
from pygame_phyics.sheet import TileSheet, SurfaceSheet

class Scene:
    """여기로 오브젝트가 등록되고 공통함수를 실행합니다
    """
    
    def __init__(self):

        self.layers = [[],[],[],[],[],[],[],[]]
    
    def layer_loop(self, method: str, *args, **kwargs):
        """레이어를 순환하며 겍체에 method 를 실행합니다

        Args:
            method (str): 실행할 함수 이름 
        """
        for layer in self.layers:
            for obj in layer:
                if kwargs.get("only") == "phyics":
                    if isinstance(obj, Physics):
                        func = getattr(obj, method)
                        func(*args)
                elif kwargs.get('collide'):
                    if getattr(obj, "collide_enter", None) != None:
                        func = getattr(obj, method)
                        func(obj.collide_enter)
                else:
                    func = getattr(obj, method)
                    func(*args)
    
    def update(self):
        """등록된 객체에 update 함수를 실행합니다
        """
        self.layer_loop("update")
    
    def set_parent(self):
        self.layer_loop("set_parent")
    
    def on_collision_enter(self):
        """충돌 함수를 호출합니다
        """
        self.layer_loop("on_collision_enter", collide=True)
        self.layer_loop("clean_collision", only="phyics")
    
    def set_physics_location(self): 
        self.layer_loop("set_location", only='phyics')
    
    def render(self, surface: pygame.Surface):
        """등록된 객체에 render 함수를 실행합니다

        Args:
            surface (pygame.Surface): 화면
        """
        self.layer_loop("render", surface, self.camera)
    
    def add(self, obj: GameObject):
        """
        오브젝트를 추가함
        
        Args:
            obj (GameObject): GameObject 를 상속받것들
        
        """
        layer = self.layers[obj.layer]
        layer.append(obj)
    
    def absorb(self, list : List):
        """
        오브젝트 집합을 추가함
        
        Args:
            list (list): 오브젝트 리스트
        
        """
        for obj in list:
            self.add(obj)
    
    def remove(self, obj: GameObject):
        """오브젝트를 새계에서 삭제하지만 오브젝트 자체는 삭제되지않음"""
        
        try:
            self.layers[obj.layer].remove(obj)
        except ValueError:
            raise ValueError("이미 삭제 되었거나 다른 레이어 계층에 있는것 같습니다")
            
    def clear(self):
        """모든 오브젝트를 새계에서 삭제하고 오브젝트 객체도 삭제됨 (delete 함수 실행)
        """
        for layer in self.layers:
            for _ in range(len(layer)):
                layer[0].delete()
    
    def get_objects(self, obj_name):
        """
        입력받은 이름을 가진 오브젝트들을 반환함
        
        Args:
            obj_name (str): 오브젝트 이름
            
        Returns:
            list: 오브젝트들
        """
        objs = []
        for layer in self.layers:
            for i in range(len(layer)):
                if layer[i].name == obj_name:
                    objs.append(layer[i])
        return objs
    
    @property
    def display(self):
        return self.__display
    
    @display.setter
    def display(self, value):
        self.__display = value
        self.camera = self.get_objects(value)[0]

    def __del__(self):
        self.clear()
        del self.layers
    
    def darkening(self):
        """
        화면 전환시 실행돠는 함수
        따로 override 해서 사용하세요
        """
        pass
    
    def brightening(self):
        """
        화면 전환시 실행돠는 함수
        따로 override 해서 사용하세요
        """
        pass

    def load(self, path: str):
        """오브젝트를 생성하고 새계에 등록합니다

        Args:
            path (str): 멥 파일에 경로 .json

        Raises:
            ImportError: path 경로에 json 에서 class_list 에 존재하지 않는 클래스를 불러오려 할때
            ValueError: 매개변수와 json 에서 저장된 값에 이름이 다르면
        """
        json : dict = _util.jsopen(path)
        setting : dict = json['setting']
        Manger.tile_sheet = {til[0] : TileSheet(*til) for til in setting['tile']}
        Manger.surface_sheet = {suf[0] : SurfaceSheet(*suf) for suf in setting['surface']}
        for key, value in setting.items():
            setattr(Manger, key, value)
        
        objs = json['objs']
        for name in objs.keys():
            for json_object in objs[name]:
                try:
                    args = list(json_object.values())
                    parameters = list(json_object.keys())
                    prefab_class = Manger.classes.get(name)
                    if prefab_class == None:
                        prefab_class = globals()[name]
                        if prefab_class == None:
                            raise ImportError(f"{name} 클레스가 존재하지 않거나 불러지지 않았습니다. \n 현재 불러온 클래스 {Manger.classes}")
                    if list(inspect.signature(prefab_class).parameters.keys()) == parameters:
                        prefab = prefab_class(*args)
                        prefab.instantiate()
                    elif len(list(inspect.signature(prefab_class).parameters.keys())) == len(parameters):
                        raise ValueError(f"리스트에 길이는 같으나 이름이 틀리거나 순서가 다른것같습니다.\njson :{parameters}\n{name} class:{list(inspect.signature(prefab_class).parameters.keys())}")
                except ValueError as e:
                    print("Error message: ", e)
        

        self.set_parent()
        
        self.set_physics_location()

        self.display = setting.get('display', 'main_cam')


import pygame
import Box2D
from pygame_phyics import PPM
from pygame_phyics.location import Location, PhysicsLocation
from pygame_phyics.manger import Manger
from pygame_phyics.objects.gameobject import GameObject
from pygame_phyics.objects.joint import Joint
from pygame_phyics.vector import Vector


class Physics(GameObject, Joint):
    """물리오브젝트에 공통점"""
    
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name):
        GameObject.__init__(self, name, layer, tag, visible, position, rotation, parent_name)
        self.collide_enter = None
        self.bpos = self.location.position
        self.brot = self.location.rotation
        self.location = PhysicsLocation(self)
        
    def set_location(self):
        self.location.position = self.bpos
        self.location.rotation = self.brot
        del self.bpos
        del self.brot
    
    def on_collision_enter(self, collision: GameObject):
        """물리 오브젝트가 충돌시 이 함수가 호출됩니다

        Args:
            collision (GameObject): 충돌하는 상대 오브젝트 입니다
        """
        pass
    
    def clean_collision(self):
        """충돌하고 상대 오브젝트 주소를 삭제해 다음 충돌을 받을수 있도록 합니다"""
        self.collide_enter = None
    
    def render(self, surface, camera):
        if self.collide_visible:
            for fixture in self.body.fixtures:
                fixture.shape.render(self.body, surface, camera)
    
    def delete(self):
        """이 오브젝트를 물리 새상과 씬에서 제거합니다"""
        Manger.world.DestroyBody(self.body)
        GameObject.delete(self)
        
        
def circle_render(circle, body, surface, camera):
    position = body.transform * circle.pos * PPM
    position = camera((position[0], Manger.HEIGHT - position[1]))
    pygame.draw.circle(surface, (0, 0, 0, 255), [int(
        x) for x in position], int(circle.radius * PPM), 1)
Box2D.b2CircleShape.render = circle_render
def polygon_render(polygon, body, surface, camera):
    vertices = [(body.transform * v) * PPM for v in polygon.vertices]
    vertices = [(v[0], Manger.HEIGHT - v[1]) for v in vertices]
    vertices = [camera(v) for v in vertices]
    pygame.draw.polygon(surface, (0, 0, 0, 255), vertices, 1)
Box2D.b2PolygonShape.render = polygon_render

class StaticObject(Physics): 
    """정적 물리 오브젝트"""
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, scale, shape_type, collide_visible):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.collide_visible = collide_visible
        match shape_type:
            case "chain":
                self.shape = Box2D.b2ChainShape(vertices_chain=scale)
            case "circle":
                self.shape = Box2D.b2CircleShape(radius=scale)
            case "edge":
                self.shape = Box2D.b2EdgeShape(vertices=scale)
            case "polygon":
                self.shape = Box2D.b2PolygonShape(vertices=scale)
        self.body : Box2D.b2Body = Manger.world.CreateStaticBody(
            shapes=self.shape
            )
        self.body.userData = self


class DynamicObject(Physics):
    """동적 물리 오브젝트"""
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, scale : tuple | float, shape_type, collide_visible,
        density, friction):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.collide_visible = collide_visible
        self.body : Box2D.b2Body = Manger.world.CreateDynamicBody()
        self.body.userData = self
        match shape_type:
            case "chain":
                self.shape = self.body.CreateChainFixture(vertices_chain=scale, density=density, friction=friction)
            case "circle":
                self.shape = self.body.CreateCircleFixture(radius=scale, density=density, friction=friction)
            case "edge":
                self.shape = self.body.CreateEdgeFixture(vertices=scale, density=density, friction=friction)
            case "polygon":
                self.fixture = self.body.CreatePolygonFixture(vertices=scale, density=density, friction=friction)


from pygame_phyics.objects.ui.ui import UI
from pygame_phyics.event import Event
from pygame_phyics.input import Input
from pygame_phyics.objects.image import ImageObject


class Button(UI):
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, default, clicked):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.default = ImageObject(self, default, follow=True, collide=True)
        self.clicked = ImageObject(self, clicked, follow=True, collide=True)
        self.image = self.default
        self.rect = self.image.rect
        self.is_click = Event()
    
    def on_mouse_stay(self, pos):
        if Input.get_mouse_down(0):
            self.is_click.invoke() 
            self.image = self.clicked
        elif Input.get_mouse_up(0):
            self.image = self.default
    
    def update(self):
        self.image.update()
        
    def render(self, surface, camera):
        self.image.render(surface, camera)



from pygame_phyics.objects.component import Component
from pygame_phyics.objects.image import ImageObject
from pygame_phyics.objects.object import Object
from pygame_phyics.objects.camera import CameraObject
from pygame_phyics.objects.parent import ParentObject
from pygame_phyics.objects.gameobject import GameObject
from pygame_phyics.objects.tilemap import TileMap
from pygame_phyics.objects.joint import Joint
from pygame_phyics.objects.physics import Physics, StaticObject, DynamicObject
from pygame_phyics.objects.ui import *




from pygame_phyics.objects.object import Object
from pygame_phyics.location import Parent


class ParentObject(Object):
    def __init__(self):
        super().__init__("parent", 0, "top")
        self.location = Parent()
        
    def set_parent(self):
        pass


import pygame
from pygame_phyics.objects.component import Component
from pygame_phyics.vector import Vector
from pygame_phyics.manger import Manger

class ImageObject(Component):
    """이미지에 위치, 각도를 관리하고 화면에 나타낸다
    씬에서 그려지는 오브젝트가 아닙니다
    """
    
    def __init__(self, object, image=None, position=(0, 0), angle=0, **kwargs):
        self.object = object
        self.visible = True
        self.og_image = pygame.image.load(image) if kwargs.get('surface') == None else pygame.Surface(kwargs['surface'], pygame.SRCALPHA)
        self.collide = kwargs.get("collide", False)
        self.is_follow_camera = kwargs.get("follow", False)
        self.image = self.og_image
        self.rect = self.image.get_rect()
        self.position = Vector(*position)
        self.angle = angle
        self.type = 'center' if kwargs.get('type') == None else kwargs['type']
    
    def update(self):
        obj_self = list(zip(self.object.render_position.xy, self.position.xy))
        obj_self_ys = obj_self[1]
        
        world_position = Vector(sum(obj_self[0]), obj_self_ys[0] - obj_self_ys[1])
        
        position = self.object.render_position.rotate_vector(world_position, self.object.location.rotation)

        self.rect = self.image.get_rect(**{self.type: position.xy})
        
        if self.collide:
            self.object.rect = self.rect


    def render(self, surface, camera):
        if self.visible:
            if self.is_follow_camera:
                self.image = pygame.transform.rotate(self.og_image, self.angle + self.object.location.rotation)
            else:
                self.image = pygame.transform.rotate(self.og_image, self.angle + self.object.location.rotation + Manger.scene.camera.location.rotation)
            
            if self.is_follow_camera:
                surface.blit(self.image, self.rect.topleft)
            else:
                surface.blit(self.image, self.image.get_rect(center=camera(self.rect.center)))




class ImmutableAttributeError(Exception):
    """불변의 값을 수정하려 시도할때 (불변은 변하지 않는)"""
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
    
class JsonSerializableError(Exception):
    """값이 json 으로 파싱할수 없는 값일때"""
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
    
class FunctionError(Exception):
    """주어진 값이 함수가 아닐떼"""
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
    
class GameOver(Exception):
    """한 맵의 루프가 끝났다는 뜻"""
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


import pygame
from pygame_phyics import game, util
from pygame_phyics.event import Event
from pygame_phyics.input import Input
from pygame_phyics.objects.image import ImageObject
from pygame_phyics.objects.ui import NEWLINE
from pygame_phyics.objects.ui.text import Text
from pygame_phyics.objects.ui.ui import UI
from pygame_phyics.timertask import OnceTimerTask, TimerTask


class InputField(UI):
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, scale, color, font, interval, rect):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.image = ImageObject(self, surface=rect, type='topleft', follow=True, collide=True)
        self.image.og_image.fill((0, 0, 0, 128))
        self.input_line = ImageObject(self, surface=(5, scale[1]), type="topleft", follow=True)
        self.input_line.og_image.fill((255,255,255, 255))
        self.field = Text(name + '_text', layer, tag + "_text", visible, [0, 0], 0, name, scale[1], color, font, interval)
        self.text = ""
        self.focused = False
        self.editing_pos = 0
        
        self.text_edit = False
        self.text_editing = ""
        self.text_editing_pos = 0
        
        self.backspace = False
        self.stay = False
        self.timertask = TimerTask(600, self.toggle_bar)
        self.backtime = TimerTask(40, lambda :self.focus_cut(1))
        self.wait_backspace = OnceTimerTask(350, self.toggle_backspace)
        
        self.input_event = Event()
        game.event_event.add_lisner(self.inputfield_event)
        
    def bar_reset(self):
        """커서에 깜빡거림에 주기를 초기화합니다
        """
        self.timertask.reset()
        self.input_line.visible = True
        
    def toggle_bar(self):
        """커서를 토글로 껏다킵니다
        """
        self.input_line.visible = not self.input_line.visible
    
    def toggle_backspace(self):
        """연속 지우기를 토글로 제어합니다
        """
        self.backspace = not self.backspace
    
    def insert(self, index: int, value: str):
        """index 위치에 value 를 삽입합니다

        Args:
            index (int): 위치
            value (str): 삽입할 글자
        """
        self.text = util.string_insert(self.text, value, index)
    
    def cut(self, range: tuple[int, int]):
        """일정 범위에 글자를 잘라냅니다

        Args:
            range (tuple[int, int]): 잘라낼 범위
        """
        self.text = util.string_cut(self.text, range)
    
    def focus_insert(self, value: str):
        """커서를 기준으로 글자를 삽입합니다

        Args:
            value (str): 삽입할 글자
        """
        self.insert(self.editing_pos, value)
        self.set_edit_pos(len(value), add=True)
    
    def focus_cut(self, size: int):
        """커서를 기준으로 글자를 잘라냅니다

        Args:
            size (int): 커서로부터 이만큼 잘라냅니다
        """
        self.cut((self.editing_pos-size, self.editing_pos))
        self.set_edit_pos(size, sub=True)
        
    def set_edit_pos(self, pos: int, **kwargs):
        """커서에 위치를 변경합니다
        add 키워드는 bool 로 True 일떄 인수로 받은 pos를 더합니다
        sub 키워드는 bool 로 True 일떄 인수로 받은 pos를 뺍니다

        Args:
            pos (int): 커서 위치, 또는 연산할 값
        """
        if kwargs.get("add"):
            pos += self.editing_pos
        elif kwargs.get("sub"):
            pos = self.editing_pos - pos
        length = len(self.text+self.text_editing)
        if 0 >= pos:
            self.editing_pos = 0
        elif pos > length:
            self.editing_pos = length
        else:
            self.editing_pos = pos
    
    def on_mouse_enter(self, pos):
        self.stay = True
    
    def on_mouse_stay(self, pos):
        if Input.get_mouse_down(0):
            self.focused = True
    
    def on_mouse_exit(self, pos):
        self.stay = False
    
    def update(self):
        if self.focused:
            if Input.get_key_down(pygame.K_BACKSPACE):
                self.wait_backspace.reset()
                if len(self.text) > 0 and self.editing_pos > 0:
                    self.focus_cut(1)
                    self.bar_reset()   

            elif Input.get_key_down(pygame.K_DELETE):
                self.cut((self.editing_pos, self.editing_pos+1))

            elif Input.get_key_down(pygame.K_LEFT):
                self.set_edit_pos(1, sub=True)
                self.bar_reset()  

            elif Input.get_key_down(pygame.K_RIGHT):
                self.set_edit_pos(1, add=True)
                self.bar_reset()  

            elif Input.get_key_down(13):
                self.focus_insert(NEWLINE)

            elif Input.get_key_down(pygame.K_KP_ENTER):
                self.focused = False
                self.input_event.invoke(self.text)

            elif Input.get_key(pygame.K_BACKSPACE):
                self.wait_backspace.run_periodic_task()

            elif Input.get_key_up(pygame.K_BACKSPACE):
                self.backspace = False

            if self.backspace:
                self.backtime.run_periodic_task()

        self.image.update()
        
        if not self.stay and Input.get_mouse_down(0):
            self.focused = False
        
        self.field.text = self.text + self.text_editing
        edit_text_pos = self.editing_pos + len(self.text_editing)
        
        if self.focused:
            self.timertask.run_periodic_task()
            
            if self.input_line.visible:
                self.input_line.position = self.field.get_position(edit_text_pos)
        
        self.input_line.update()
    
    def inputfield_event(self, event):
        """입력 필드에 이벤트

        Args:
            event (event): pygame 에 이벤트입니다_
        """
        if self.focused:
            if event.type == pygame.TEXTEDITING:
                self.text_edit = True
                self.text_editing = event.text
                self.text_editing_pos = event.start
                self.bar_reset()    
            elif event.type == pygame.TEXTINPUT:
                self.text_edit = False
                self.text_editing = ""
                self.focus_insert(event.text)
                self.bar_reset()    

    def render(self, surface, camera):
        self.image.render(surface, camera)
        if self.focused:
            self.input_line.render(surface, camera)
            
    def instantiate(self):
        super().instantiate()
        self.field.instantiate()


from pygame import Surface
from pygame_phyics.camera import Camera
from pygame_phyics.objects.object import Object
from pygame_phyics.location import Location
from pygame_phyics.vector import Vector
from pygame_phyics.manger import Manger
from pygame_phyics.effect import Glitch

class CameraObject(Object):
    def __init__(self, name, tag, position, rotation):
        super().__init__(name, 0, tag)
        self.visible = False
        self.status = "idle"
        self.rect = None
        self.collide = 'out'
        self.glitch = Glitch()
        self.location = Location(Vector(*position), rotation)

    def set_parent(self):
        parent = Manger.scene.get_objects("parent")[0]
        self.location.set_parent(parent.location)

    def __call__(self, position: tuple[int, int] | Vector):
        """카메라 시선을 적용한 위치를 반환함

        Args:
            position (tuple[float, float] | Vector): 오브젝트에 위치

        Returns:
            tuple[float, float]: 카메라 시선이 적용된 위치
        """
        rotated = Vector(Manger.WIDTH/2, Manger.HEIGHT/2).rotate_vector(Vector(*position), self.location.rotation)
        camerad = rotated - self.location.position
        return camerad.xy

    def render(self, surface: Surface, camera: Camera):
        surface.fill((127, 127, 127))


import inspect
from pygame import time
from typing import Callable
from pygame_phyics.error import FunctionError

class TimerTask:
    """틱마다 이벤트를 실행합니다
    **주의사항 int 나 str 같은것들은 생성 당시 값을 가지므로
    실시간 숫자를 가져오려면 람다식을 사용하시는게 좋습니다
    """
    def __init__(self, tick: int, event: Callable = lambda:None, *value, **kwargs):
        """틱마다 이벤트를 실행합니다
        **주의사항 int 나 str 같은것들은 생성 당시 값을 가지므로
        실시간 숫자를 가져오려면 람다식을 사용하시는게 좋습니다
        
        Args:
            tick (int): 실행 간격
            event (Callable): 실행할 함수

        Raises:
            FunctionError: 주어진 event 가 함수가 아닐떄
        """
        self.tick = tick
        self.last_update = 0
        if inspect.ismethod(event) and inspect.isfunction(event):
            raise FunctionError(f"입력받은 값은 는 함수가 아닙니다")
        self.event = event
        self.value = value
        self.kwargs = kwargs
        
    def run_periodic_task(self):
        """요구 시간이 지났는지 확인하는 함수
        확인될시 True 를 반환한다

        Returns:
            _type_: _description_
        """
        if time.get_ticks() - self.last_update > self.tick:
            self.last_update = time.get_ticks()
            self.event(*self.value, **self.kwargs)
            return True
        return False
    
    def reset(self):
        """기다리는것을 처음으로 되돌린다 생각하면 됩니다
        재생중인 영상을 처음으로 이동하는것처럼
        """
        self.last_update = time.get_ticks()
        
class OnceTimerTask(TimerTask):
    """한번 실행되는 TimerTask 
    리셋되멘 다시 실행할수 있다
    기다릴떄 리셋하고 실행하는게 좋을것같다
    """
    def __init__(self, tick, event=lambda:None, *value, **kwargs):
        super().__init__(tick, event, *value, **kwargs)
        self.once = False
    
    def run_periodic_task(self):
        if time.get_ticks() - self.last_update > self.tick and not self.once:
            self.event(*self.value, **self.kwargs)
            self.once = True
            return True
        return False
    
    def reset(self):
        super().reset()
        self.once = False
    


NEWLINE = '\n'

from pygame_phyics.objects.ui.ui import UI
from pygame_phyics.objects.ui.button import Button
from pygame_phyics.objects.ui.inputfield import InputField
from pygame_phyics.objects.ui.text import Text


"""

게임 모듈 

Example:
    게임의 초기 샛팅을 실행하고 
    게임 맵을 정의하고
    실행한다
    literal blocks::

        $ Game.init((1000, 800), "MyGame")
        $
        $ @pygame_phyics.world("example/example.json")
        $ def main():
        $   Game.loop([], [])
        $   
        $ main()

"""

import pygame
from pygame import display

from Box2D import b2World, b2ContactListener
from typing import Callable
from pygame_phyics.manger import Manger
from pygame_phyics.scene import Scene
from pygame_phyics.mouse import mouse_event
from pygame_phyics.instantiate import import_module
from pygame_phyics.input import Input
from pygame_phyics.event import Event
import pygame_phyics.mouse as mouse


class ContactListener(b2ContactListener):
    """물리 새상에서 충돌 연산을 GameObject 로 꺼낼수 있게 도와줌"""
    def BeginContact(self, contact):
        fixtureA = contact.fixtureA
        fixtureB = contact.fixtureB
        a_obj = fixtureA.body.userData 
        b_obj = fixtureB.body.userData
        if a_obj != None:
            a_obj.collide_enter = b_obj
        if b_obj != None:
            b_obj.collide_enter = a_obj

def world(world_path: str):
    """함수를 새상으로 등록하는 데코레이터 입니다
    
    Args:
        world_path (str) : json 경로
    """
    def real_world(func):
        def wrapper():
            Manger.scene.darkening()
            Manger.world = b2World()
            contact_listener = ContactListener()
            Manger.world.contactListener = contact_listener
            Manger.scene = Scene()
            Manger.scene.load(world_path)
            start, event, update = func()
            start(Game)
            Manger.scene.brightening()
            Game.loop(event, update)
            del Manger.scene
            del Manger.world
        return wrapper
    return real_world

event_event = Event()

class Game:
    """게임 엔진"""
    is_running = True
    clock = pygame.time.Clock()
    time_step = 1 / 60
    ve_iters = 10
    pos_iters = 10
    
    @classmethod
    def init(cls, size : tuple[int, int], title : str):
        """게임에 초기 설정을 진행합니다

        Args:
            size (tuple[int, int]): 스크린 크기입니다
            title (str): 프로그램에 제목입니다
        """
        pygame.init()
        display.set_caption(title)
        pygame.key.start_text_input()
        Manger.init(display.set_mode(size), Scene())
    
    @classmethod
    def import_objects(cls, obj_dir : str, **kwargs):
        """클래스들을 불러와 Manger 에 저장합니다

        Args:
            obj_dir (str): 오브젝트 클래스에 파일를 저장한 폴더 경로
        """
        Manger.classes.update(import_module(obj_dir, **kwargs))
    
    @classmethod
    def stop(cls):
        cls.is_running = False
    
    @classmethod
    def loop(cls, events: Callable, func: Callable):
        """이벤트랑 함수를 받아 반복문 안에서 실행합니다

        Args:
            events (Callable): 이벤트 떄 실행됨
            func (Callable): 루프 돌때 실행됨
        """
        
        while cls.is_running:
            cls.clock.tick(60)
            
            func(cls)
            
            mouse_pressed = mouse.get_pressed()
            
            for i in range(3):
                mouse_click = Input.mouse_click[i]
                if mouse_pressed[i]:
                    if mouse_click <= 1:
                        Input.mouse_click[i] = 2
                    else:
                        Input.mouse_click[i] = 3
                else:
                    if mouse_click >= 2:
                        Input.mouse_click[i] = 1
                    else:
                        Input.mouse_click[i] = 0
                        
            mouse_event()
            
            for key, value in Input.key_board.items():
                if value == 2:
                    Input.key_board[key] = 3
                elif value == 1:
                    Input.key_board[key] = 0
            
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    quit()
                elif event.type == pygame.KEYDOWN:
                    Input.key_board[event.key] = 2
                elif event.type == pygame.KEYUP:
                    Input.key_board[event.key] = 1
                    
                event_event.invoke(event)
                
                events(cls, event)

            Manger.scene.update()
            
            Manger.scene.on_collision_enter()

            Manger.world.Step(
                cls.time_step,
                cls.ve_iters,
                cls.pos_iters
                )
            
            Manger.scene.render(Manger.screen)

            if Manger.scene.camera.status == 'glitch':
                Manger.scene.camera.glitch.render(Manger.screen)

            pygame.display.update()


import random
import pygame
from pygame_phyics.timertask import TimerTask

class Glitch:
    def __init__(self):
        self.timer = TimerTask(50)
        self.interval_timer = TimerTask(150)
    
    def _get_subsurface(self,surface, y):
        wid = surface.get_width()
        height = random.randrange(10, 20)
        shift = random.randint(10, 20)
        shift *= random.choice((1, -1))
        if y + height > surface.get_height():
            height = surface.get_height() - y

        sub = surface.subsurface((0, y, wid, height))
        sub_rect = sub.get_rect()
        result = pygame.Surface(sub_rect.size)

        result.blit(sub, (shift, 0))
        if shift > 0:
            sub_shift = sub.subsurface((0, 0, wid - shift, height))
            result.blit(sub_shift, (0, 0))
        else:
            sub_shift = sub.subsurface(abs(shift), 0, wid - abs(shift), height)
            result.blit(sub_shift, (wid - abs(shift), 0))
        return result

    def render(self, surface):
         if self.interval_timer.run_periodic_task():
            self.interval_timer.tick = int(random.uniform(0.2, 0.4) * 200)
            y_travelled = 0
            while y_travelled < surface.get_height():
                if random.random() > 0.8:
                    strip = self._get_subsurface(surface, y_travelled)
                    y_travelled += strip.get_height()
                    surface.blit(strip, (0, y_travelled))
                else:
                    y_travelled += random.randrange(10, 20)


"""GameObject 를 Scene 에 등록하는 모듈"""

import os
import importlib.util
import inspect

current_dir = os.path.dirname(os.path.abspath(__file__))
"""절대경로"""


def import_module(import_dir, **kwargs):
    """폴더를 선회하며 가져온 클래스를 반환함
    **주의사항 폴더 끝에 / 를 붙여야함
    
    Args:
        import_dir (str): 디랙토리 경로
        
    Returns:
        list: 클래스들
    
    """
    class_list = {}
    for file in os.listdir(import_dir):
        if file.endswith(".py") and file != '__init__.py':
            classes = import_classes(file[:-3], import_dir, **kwargs)
            class_list.update(classes)
        elif os.path.isdir(file):
            class_list += import_module(import_dir + file + "/", **kwargs)
        else:
            if kwargs.get('debug', "default") == 'detail':
                print(f"ignored file {file}")
    return class_list

def import_classes(file: str, dir: str, **kwargs) -> dict:
    """
    모듈에서 가져온 클래스를 리스트로 반환함
    
    Args:
        file (str): 파일 이름
        dir (str): 폴더 경로
    
    Returns:
        list: 클래스들    
    """
    class_list = {}
    spec = importlib.util.spec_from_file_location(file, dir + file + ".py")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    try:
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and hasattr(obj, 'instantiate'):
                if not name.startswith("_"):
                    if kwargs.get('debug', "default") == 'log':
                        print(f"Imported class: {name} from {file}.py")
                    elif kwargs.get('debug', "default") == 'detail':
                        print(f"Imported class: {name} from {file}.py\nDir: {dir + file}.py")
                    class_list[name] = obj
            else:
                if kwargs.get('debug', "default") == 'detail':
                    if not name.startswith("__"):
                        print(f"ignored: {name} from {file}.py")
    except ImportError as e:
        print(f"모듈 {file}을(를) 임포트하는 중 오류 발생: {e}")
    finally:
        return class_list



"""
pygame 의 마우스르 상속받음
"""
from pygame_phyics.manger import Manger
from pygame.mouse import *

def mouse_event():
    """rect 를 가진 오브젝트의 마우스 충돌을 연산함

    Raises:
        NotImplementedError: 마우스 충돌은 가장 위 레이어 에만 호출 되어야만 한다
        NotImplementedError: 마우스 충돌은 가장 위 레이어 에만 호출 되어야만 한다
        NotImplementedError: 마우스 충돌은 가장 위 레이어 에만 호출 되어야만 한다
    """
    
    try:
        for layer in Manger.scene.layers[::-1]:
            for obj in layer[::-1]:
                if obj.rect != None:
                    if obj.rect.collidepoint(get_pos()):
                        if obj.collide == 'out':
                            obj.collide = 'in'
                            obj.on_mouse_enter(get_pos())
                            raise NotImplementedError
                        elif obj.collide == 'in':
                            obj.on_mouse_stay(get_pos())
                            raise NotImplementedError
                    elif obj.collide == 'in':
                        obj.collide = 'out'
                        obj.on_mouse_exit(get_pos())
                        raise NotImplementedError
    except:
        pass


from pygame_phyics import util
from pygame_phyics.location import Location
from pygame_phyics.manger import Manger
from pygame_phyics.objects.object import Object
from pygame_phyics.vector import Vector


class GameObject(Object):
    """마우스 충돌 연산가능함"""
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name):
        super().__init__(name, layer, tag)
        self.visible = visible
        self.rect = None
        self.collide = 'out'
        self.location = Location(Vector(*position), rotation)
        self.parent_name = parent_name
    
    @util.getter
    def render_position(self):
        """y 좌표를 반전시켜줍니다
        pygame 은 화면에 그릴떄 y 선이 아래로 이동할수록 + 라서 
        사용자가 햇갈릴수 있기 떄문에 멥 파일에선 우리가 사용하는 방식으로작성하고
        그릴떄 좌표를 이것으로 사용하시면 됩니다.
        하지만 좌표가 topleft 기준입니다"""
        return Vector(self.location.world_position.x, Manger.HEIGHT - self.location.world_position.y)
    
    def set_parent(self):
        parent: GameObject = Manger.scene.get_objects(self.parent_name)[0]
        self.location.set_parent(parent.location)


import inspect
import pygame
from pygame_phyics.util import jsave
from pygame_phyics.manger import Manger

def save(export_path: str, **kwargs):
    """현재 맵을 저장해 export_path 경로로 저장합니다

    Args:
        export_path (str): 내보낼 경로

    Raises:
        ValueError: 저장할 겍체가 메게변수와 같은 변수가 없습니다
    """
    kwargs['camera'] = list(Manger.scene.camera.vector.xy + (Manger.scene.camera.angle, ))
    saved_dict = {
        'setting':kwargs,
        'objs':{
            
        }
    }

    objs = {}
    for layer in Manger.scene.layers:
        for obj in layer:
            prargs = {}
            parameters = list(inspect.signature(obj.__class__).parameters.keys())
            for parameter in parameters:
                value = getattr(obj, parameter, None)
                if value == None:
                    raise ValueError(f"{obj.__class__.__name__}는 {parameter} 속성이 존재하지 않습니다")
                if hasattr(value, "isiter"):
                    prargs[parameter] = list(value)
                elif isinstance(value, pygame.rect.Rect):
                    prargs[parameter] = [value.width, value.height]
                else:
                    prargs[parameter] = value
            class_name = obj.__class__.__name__
            ishas_class = objs.get(class_name)
            if ishas_class == None:
                objs[class_name] = [prargs]
            else:
                objs[class_name].append(prargs)
    saved_dict['objs'] = objs
    
    jsave(saved_dict, export_path)


import pygame
from pygame_phyics.camera import Camera

class Component:
    """기본 함수만 선언해놓음"""
    def on_mouse_enter(self, pos: tuple[int, int]):
        """마우스가 오브젝트(rect) 안에 들어올떄 호출됨

        Args:
            pos (tuple[int, int]): 마우스 위치 (접촉한 위치)
        """
        pass
    def on_mouse_stay(self, pos: tuple[int, int]):
        """마우스가 오브젝트(rect) 에서 들어온 상태일떄 계속 호출됨

        Args:
            pos (tuple[int, int]): 마우스 위치 (접촉한 위치)
        """
        pass
    def on_mouse_exit(self, pos: tuple[int, int]):
        """마우스가 오브젝트(rect) 에서 들어온 상태였다가 나오면 호출됨

        Args:
            pos (tuple[int, int]): 마우스 위치 (접촉한 위치)
        """
        pass
    def update(self):
        """게임 루프시 한번씩 실행됩니다"""
        pass
    def render(self, surface: pygame.Surface, camera: Camera):
        """씬에서 그릴때 실행합니다

        Args:
            surface (pygame.Surface): 프로그램 화면
            camera (Camera): 씬 카메라
        """
        pass
    


from typing import List, Callable

class Event:
    """이 클래스는 함수들을 받아 특정상황에 실행합니다
    
    Attributes:
        lisners: A list include functions
    """
    def __init__(self):
        """이벤트를 생성한다"""
        self.lisners : List[Callable] = []
    
    def __call__(self):
        """이벤트를 호출한다"""
        self.invoke()
    
    def __len__(self):
        """이벤트 안에 등록된 함수의 갯수를 반환함

        Returns:
            int: 함수의 갯수
        """
        return len(self.lisners)
    
    def __iadd__(self, value):
        self.add_lisner(value)
    
    def __isub__(self, value):
        self.remove(value)
    
    def add_lisner(self, function: Callable):
        """함수를 인자로 받아 함수를 등록합니다.

        Args:
            function (Callable): 등록할 함수
        """
        self.lisners.append(function)
    
    def remove(self, function: Callable):
        """함수를 리스트에서 제거합니다

        Args:
            function (Callable): 제거할 함수 자체
        """
        self.lisners.remove(function)
    
    def clear(self):
        self.lisners.clear()

    def invoke(self, *arg):
        """lisners의 함수들을 실행합니다."""
        for function in self.lisners:
            function(*arg)


import pygame
import re

# 내 모듈 임포트 못하는 모듈
class Manger:
    """정적 변수를 이곳에 저장함"""
    
    classes = {}
    obj_names = []

    @classmethod
    def check_object_name(cls, name: str):
        for obn in cls.obj_names:
            if obn == name:
                intr: str = re.compile('\(([^)]+)').findall(name)[-1]
                if intr and intr.isdigit():
                    return cls.check_object_name(name.replace(f"({intr})", f"({int(intr) + 1})"))
                return name + " (1)"
            else:
                continue
        return name

    @classmethod
    def init(cls, screen: pygame.Surface, none_scene):
        """게임에 정적 변수 초기 설정 함수

        Args:
            screen (pygame.Surface): 프로그램 스크린
            none_scene (Scene): 공백인 씬
        """
        cls.screen = screen
        cls.scene = none_scene
        cls.WIDTH, cls.HEIGHT = screen.get_size()


from pygame_phyics.objects.gameobject import GameObject
from pygame_phyics.vector import Vector


class UI(GameObject):
    """ui 를 위한 오브젝트"""
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)


from pygame_phyics.manger import Manger

class Joint:
    """물리 연산의 연결 담당"""
    joints = []
    
    def create_joint(self, frequency_hz, damping_ratio, body):
        """자기 자신과 인수로 받은 body 를 연결합니다"""
        joint = Manger.world.CreateDistanceJoint(
            frequencyHz=frequency_hz,
            dampingRatio=damping_ratio,
            bodyA=self.body,
            bodyB=body.body,
            anchorA=self.body.transform.position,
            anchorB=body.body.transform.position
        )
        Joint.joints.append(joint)
        return joint


from pygame_phyics import Manger
from pygame_phyics.timertask import TimerTask

class Animation:
    def __init__(self, tick, image_object, **kwargs):
        self.object = image_object
        self.period = TimerTask(tick)
        self.index = 0
        if  kwargs.get('sheet', None) != None:
            self.sheet = Manger.surface_sheet[kwargs['sheet']]
            if kwargs.get('range', None) != None:
                frist, last = kwargs['range']
                self.images = self.sheet.images[frist:last]
            else:
                raise ValueError("범위를 설정해야함 (int, int) 튜플형식으로 재공")
        else:
            raise ValueError("키워드가 적절하지 않습니다. sheet 를 사용하면 sheet 키워드에 사용하는 시트에 이름을 작성하세요, xml 문서를 이용한 이미지의 집합을 사용하시면 \'xml\'를 작성하세요")
        self.len = len(self.images)
        
    def update(self):
        if self.period.run_periodic_task():
            self.change_image()

    def change_image(self):
        self.object.og_image = self.images[self.index]
        self.index += 1
        if self.index == self.len:
            self.index = 0


from pygame_phyics.manger import Manger
from pygame_phyics.objects.component import Component


class Object(Component):
    """
    새상에 등록할수있는 가장 기초적인 오브젝트
    save 사용시 주의사항: 매개변수 이름이랑 저장용 변수 이름이랑 이름이 같아야합니다
    """
    
    def __init__(self, name, layer, tag):
        self.name = Manger.check_object_name(name)
        self.tag = tag
        self.layer = layer

    def delete(self): 
        """씬에서 이 오브젝트를 삭제합니다"""
        Manger.scene.remove(self)
        del self

    def instantiate(self):
        Manger.scene.add(self)



__version__ = "1.1.2"

PPM =20

from pygame_phyics.manger import Manger
from pygame_phyics.game import Game
from pygame_phyics.objects import *
from pygame_phyics.sheet import SurfaceSheet
from pygame_phyics.vector import Vector
from pygame_phyics.timertask import TimerTask, OnceTimerTask
from pygame_phyics.input import Input
from pygame_phyics.animation import Animation

from pygame import (
    Surface,
    Rect
)

from pygame.constants import *



import math
from pygame_phyics.util import const

class Vector:
    def __init__(self, x: float, y:float):
        self.x = x
        self.y = y
        self.index = 0
    
    def __getitem__(self, index):
        match index:
            case 0:
                return self.x
            case 1:
                return self.y

    def isiter():pass

    def __iter__(self):
        return self
    
    def __next__(self):
        if self.index < 2:
            item = (self.x, self.y)[self.index]
            self.index += 1
            return item
        else:
            self.index = 0
            raise StopIteration
    
    
    @const
    def xy(self):
        return self.x, self.y
    
    @const
    def empty(self):
        return Vector(0, 0)
    
    def rotate_vector(self, vector, angle):
        angle_radians = math.radians(360 - angle)
        x_a, y_a = self.xy
        x_b, y_b = vector.xy
        x = (x_b - x_a) * math.cos(angle_radians) - (y_b - y_a) * math.sin(angle_radians) + x_a
        y = (x_b - x_a) * math.sin(angle_radians) + (y_b - y_a) * math.cos(angle_radians) + y_a
        return Vector(x, y)
    
    def add_vector(self, vector):
        x= self.x + vector.x
        y= self.y + vector.y
        return Vector(x,y) 
    
    def sub_vector(self, vector):
        x= self.x - vector.x
        y= self.y - vector.y
        return Vector(x,y) 
        
    def div_vector(self, vector):
        x= self.x / vector.x
        y= self.y / vector.y
        return Vector(x,y) 
        
    def mul_vector(self, vector):
        x= self.x * vector.x
        y= self.y * vector.y
        return Vector(x,y) 
    
    def add_float(self, float):
        x= self.x + float
        y= self.y + float
        return Vector(x,y) 
        
    def sub_float(self, float):
        x= self.x - float
        y= self.y - float
        return Vector(x,y) 
        
    def div_float(self, float):
        x= self.x / float
        y= self.y / float
        return Vector(x,y) 
    
    def mul_float(self, float):
        x= self.x * float
        y= self.y * float
        return Vector(x,y) 
        
    def __add__(self, value):
        if type(value) == Vector:
            return self.add_vector(value)
        elif type(value) in [int, float]:
            return self.add_float(value)
        else:
            raise TypeError(f"{type(value)} 는 알맞은 타입이 아닙니다 Vector or float/int")
    
    def __sub__(self, value):
        if type(value) == Vector:
            return self.sub_vector(value)
        elif type(value) in [int, float]:
            return self.sub_float(value)
        else:
            raise TypeError(f"{type(value)} 는 알맞은 타입이 아닙니다 Vector or float/int")
    
    def __div__(self, value):
        if type(value) == Vector:
            return self.div_vector(value)
        elif type(value) in [int, float]:
            return self.div_float(value)
        else:
            raise TypeError(f"{type(value)} 는 알맞은 타입이 아닙니다 Vector or float/int")
    
    def __mul__(self, value):
        if type(value) == Vector:
            return self.mul_vector(value)
        elif type(value) in [int, float]:
            return self.mul_float(value)
        else:
            raise TypeError(f"{type(value)} 는 알맞은 타입이 아닙니다 Vector or float/int")


from collections.abc import Callable
import math
import json as jsmodule
from pygame_phyics.error import JsonSerializableError
from pygame_phyics.error import ImmutableAttributeError

class const(property):
    """변수를 불변으로 만들고 싶을떄 프로퍼티처럼 사용하면 됩니다
    """
    def settar(self, value):
        raise ImmutableAttributeError("이 속성은 불변입니다 값을 할당하려 하지 마십시오")

class getter(property):
    """변수를 수정하기 싫을떄 프로퍼티처럼 사용하면 됩니다
    const 클래스와 다른점은 없지만 내부적으로 변수가 변할시 불변이랑 거리가 멀어서 새로 만들었습니다
    """
    def settar(self, value):
        raise ImmutableAttributeError("이 속성은 할당할수없습니다 값을 할당하려 하지 마십시오")

def string_insert(string: str, insert_string: str, index: int):
    """문자열에 문자열을 삽입합니다

    Args:
        string (str): 문자열
        insert_string (str): 삽입할 문자열
        index (int): 위치

    Raises:
        ValueError: 위치가 잘못된 값일떄

    Returns:
        str: 삽입된 문자열 
    """
    if len(string) < index or index < 0:
        raise ValueError(f"index 가 문자열에 유효한 위치가 아닙니다 string len: {len(string)} index: {index}")
    return f"{string[:index]}{insert_string}{string[index:]}"

def string_cut(string: str, range: tuple[int, int]):
    """일정 범위에 문자열을 잘라냅니다

    Args:
        string (str): 잘릴 문자열
        range (tuple[int, int]): 범위

    Raises:
        ValueError: 범위가 오름차순이 아닐떄

    Returns:
        _type_: _description_
    """
    if range[0] > range[1]:
        raise ValueError("범위에 끝은 처음보다 크거나 같아야합니다")
    return f"{string[:range[0]]}{string[range[1]:]}"

def jsopen(path: str) -> dict:
    """json 파일을 열어 dict 로 파싱합니다

    Args:
        path (str): json 경로

    Returns:
        dict: 파싱된 dict
    """
    with open(path, 'r') as f:
        json = jsmodule.loads(f.read())
    return json

def jsave(data: dict, path: str):
    """

    Args:
        data (dict): 저장할 딕셔너리
        path (str): 내보낼 경로
    """
    with open(path, 'w') as f:
        jsmodule.dump(data, f, indent=4)
        
types = (str, int, float, list, bool)

def check_json_serializable(dic: dict):
    """이 딕셔너리에서 json 으로 변환할수 없는 값이 있는지 검사합니다

    Args:
        dic (dict): 검사할 딕셔너리

    Raises:
        JsonSerializableError: json 에서 저장할수 없는 타입이 발견될시
    """
    for value in list(dic.values()):
        if isinstance(value, types):
            pass
        elif isinstance(value, dict):
            check_json_serializable(value)
        else:
            raise JsonSerializableError(f"{type(value)} 타입은 json 에 저장할수 있는 타입이 아닙니다")




from setuptools import setup, find_packages

setup(
    name='pygame_phyics',
    version='1.1.2',
    author= 'fireing123',
    author_email= 'gimd82368@gmail.com',
    url= 'https://github.com/fireing123/pygame_phyics',
    install_requires=["pygame", "box2d", "box2d-py"],
    packages=find_packages(),
    long_description=open('README.md', 'r', encoding="UTF8").read(),
    long_description_content_type='text/markdown',
)



import pygame
from pygame_phyics import util
from pygame_phyics.manger import Manger
from pygame_phyics.objects.gameobject import GameObject


class TileMap(GameObject):
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, tiles, sheet_name):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.tiles = tiles
        self.sheet = Manger.tile_sheet[sheet_name]
        self.sheet_name = sheet_name
        self.size = self.sheet.size
        self.canvas = self.sheet.surfaces

    def set_tile(self, xy, value):
        x, y = xy
        layer = 0
        match xy:
            case n if n[0] >= 0 and n[1] >= 1:
                y -= 1
                layer = 0
            case n if n[0] <= -1 and n[1] >= 1:
                x = -x -1
                y -= 1
                layer = 1
            case n if n[0] < 0 and n[1] < 0:
                x = -x -1
                y = -y
                layer = 2
            case n if n[0] >= 0 and n[1] < 0:
                y = -y
                layer = 3
        try:
            ly = self.tiles[layer][y]
            l = len(ly) - 1
            if l >= x:
                ly[x] = value
            else:
                m = x -l
                ly.extend([None] * m)
                ly[x] = value
        except IndexError as e:
            ly = [None] * (n[0] + 1)
            ly = [[n[0]]] = value
            l = len(self.tiles[layer]) - 1
            if l >= y:
                self.tiles[layer][y] = ly
            else:
                m = y - l
                self.tiles[layer].extend([[]] * m)
                self.tiles[layer][y] = ly

    def get_tile(self, xy, db=None):
        try:
            match xy:
                case n if n[0] >= 0 and n[1] >= 1:
                    return self.tiles[0][n[1] - 1][n[0]]
                case n if n[0] <= -1 and n[1] >= 1:
                    return self.tiles[1][n[1] - 1][-n[0] - 1]
                case n if n[0] <= -1 and n[1] <= 0:
                    return self.tiles[2][-n[1]][-n[0] - 1]
                case n if n[0] >= 0 and n[1] <= 0:
                    return self.tiles[3][-n[1]][n[0]]
        except:
            return None
    
    def render(self, surface, camera):
        HALF_WIDTH = Manger.WIDTH / (self.size * 2)
        HALF_HEIGHT = Manger.HEIGHT / (self.size * 2)
        tile_camera = camera.location.world_position.div_float(self.size)
        xrange = int(tile_camera.x - HALF_WIDTH), int(tile_camera.x + HALF_WIDTH) + 1
        yrange = int(tile_camera.y - HALF_HEIGHT), int(tile_camera.y + HALF_HEIGHT) + 1
        pygame.draw.circle(surface, (255, 255, 255), (500, 400), 5)
        for y in range(*yrange):
            for x in range(*xrange):
                tile_n = self.get_tile((x, y))
                if tile_n != None:
                    image = self.canvas[tile_n]
                    cx = (HALF_WIDTH + x) * self.size - camera.location.world_position.x
                    cy = (HALF_HEIGHT - y) * self.size + camera.location.world_position.y
                    surface.blit(image, (cx, cy))


import pygame
from pygame_phyics.objects.ui import NEWLINE
from pygame_phyics.objects.ui.ui import UI
from pygame_phyics.vector import Vector


class Text(UI):
    """글자를 화면에 나타냄"""
    def __init__(self, name, layer, tag, visible, position, rotation, parent_name, size, color, Font, interval):
        super().__init__(name, layer, tag, visible, position, rotation, parent_name)
        self.Font = Font
        if Font.startswith('./'):
            self.font = pygame.font.Font(Font, size)
        else: 
            self.font = pygame.font.SysFont(Font, size)
        self.size = size
        self.interval = interval
        self.color = color
        self.text = ""
    
    def get_position(self, __index: int) -> Vector:
        """글자에 index 주소로 접근해서 position 을 0, 0 으로 할때 x,y 좌표를 반환합니디

        Args:
            __index (int): 글자 위치

        Returns:
            Vector: position 을 0, 0 으로 할때 x,y 좌표, 월드 좌표가 아닙니다
        """
        line = self.get_line(__index)
        text = self.text[:__index].split(NEWLINE)[-1]
        x, _ = self.font.size(text)
        y = self.size + self.interval
        y *= line
        return Vector(x-5, -y)

    def get_line(self, __index: int) -> int:
        """글자에 index 주소로 접근해서 이 글자가 어느 라인에 위치한지 찾습니다

        Args:
            __index (int): 글자위치

        Returns:
            int: 라인 번호
        """
        text = self.text[:__index]
        line = text.count(NEWLINE)
        return line

    def render(self, surface : pygame.Surface, camera):
        texts = self.text.split(NEWLINE)
        self.images = [self.font.render(text, True, self.color) for text in texts]
        positions = []
        x, y = self.render_position
        for i in texts:
            positions.append((x, y))
            y += self.size + self.interval
        surface.blits(list(zip(self.images, positions)))


from pygame import Surface
from pygame import image
import xml.etree.ElementTree as xml

class SurfaceSheet:
    def __init__(self, name, path):
        self.name = name
        root = xml.parse(path)
        self.image = image.load(root.attrib['path'])
        self.images = []
        for child in root:
            att = child.attrib
            xywh = map(int, (att['x'], att['y'], att['width'], att['height']))
            self.images.append(self.image.subsurface(xywh))
    
class TileSheet:
    def __init__(self, name, path, size):
        self.name = name
        self.size = size
        self.image = image.load(path)
        self.surfaces = []
        x, y = self.image.get_size()
        fx = 0
        fy = 0
        for i in range(int(y / size)):
            for j in range(int(x / size)):
                self.surfaces.append(self.image.subsurface((fx, fy, size, size)))
                fx += size
            fy += size
            fx = 0


from pygame_phyics.vector import Vector
from pygame_phyics.util import getter
from pygame_phyics import PPM

class Parent:
    def __init__(self):
        self.position = Vector(0, 0)
        self.world_position = Vector(0, 0)
        self.rotation = 0
        self.world_rotation = 0
        self.children = []

class Location(Parent):
    def __init__(self, position: Vector, rotation: int):
        self.position = position
        self.rotation = rotation
        

    def set_parent(self, parent: Parent):
        self.parent = parent
        parent.child = self
        
        
    @getter
    def world_position(self):
        return self.parent.world_position + self.position
    
    @getter
    def world_rotation(self):
        return self.parent.world_rotation + self.rotation
    
class PhysicsLocation:
    def __init__(self, physics):
        self.physics = physics
        self.children = []
    
    def set_parent(self, parent: Parent):
        self.parent = parent
    
    @property
    def position(self):
        return self.world_position - self.parent.world_position
    
    @position.setter
    def position(self, value: Vector):
        self.physics.body.transform.position.x = (self.parent.position.x + value.x) / PPM
        self.physics.body.transform.position.y = (self.parent.position.y + value.y) / PPM
    
    @property
    def rotation(self):
        parent = self.parent.world_rotation
        return self.physics.body.angle - parent
    
    @rotation.setter
    def rotation(self, value: int):
        self.physics.body.angle = self.parent.world_rotation + value
    
    @getter
    def world_position(self):
        return Vector(self.physics.body.transform.position.x, self.physics.body.transform.position.y) * PPM
    
    @getter
    def world_rotation(self):
        return self.physics.body.angle


"""pygame_phyics 에서 입력을 확인합니다"""

class Input:
    """키보드나 마우스의 입력을 확인합니다"""
    key_board = {}
    
    mouse_click = [0, 0, 0]
    KEYUPING = 0
    KEYUP = 1
    KEYDOWN = 2
    KEYDOWNING = 3
    @classmethod
    def get_key(cls, key_code: int) -> bool:
        """입력을 받는지 확인함니다

        Args:
            key_code (int): 파이게임 키 -> pygame.K_XXX

        Returns:
            bool: 키보드가 눌림 상태일떄 True
        """
        
        if cls.key_board.get(key_code, cls.KEYUPING) >= cls.KEYDOWN:
            return True
        return False
    
    @classmethod
    def get_key_down(cls, key_code: int) -> bool:
        """이 키가 눌리는 순간인지 확인합니다

        Args:
            key_code (int): 파이게임 키 -> pygame.K_XXX

        Returns:
            bool: 키보드를 누를떄 한번 True
        """
        
        if cls.key_board.get(key_code, cls.KEYUPING) == cls.KEYDOWN:
            return True
        return False
    
    @classmethod
    def get_key_up(cls, key_code: int) -> bool:
        """이 키가 더이상 입력을 받지 않는지 확인합니다

        Args:
            key_code (int): 파이게임 키 -> pygame.K_XXX

        Returns:
            bool: 키보드를 눌렀다가 떗을떄 한번 True
        """
        
        if cls.key_board.get(key_code, cls.KEYUPING) == cls.KEYUP:
            return True
        return False
    
    @classmethod
    def get_mouse(cls, button: int) -> bool:
        """클릭중인지 확인합니다

        Args:
            button (int): 마우스 버튼 0: 왼쪽, 1: 휠, 2:오른쪽 버튼 입니다

        Returns:
            bool: 마우스가 눌림 상태일떄 True
        """
        
        if cls.mouse_click[button] <= 2:
            return True
        return False
    
    @classmethod
    def get_mouse_down(cls, button: int) -> bool:
        """클릭 하는 순간인지 확인합니다

        Args:
            button (int): 마우스 버튼 0: 왼쪽, 1: 휠, 2:오른쪽 버튼 입니다

        Returns:
            bool: 마우스가 눌릴떄 한번 True
        """
        
        if cls.mouse_click[button] == 2:
            return True
        return False
    
    @classmethod
    def get_mouse_up(cls, button: int) -> bool:
        """더이상 클릭을 하지 않는지 확인합니다

        Args:
            button (int): 마우스 버튼 0: 왼쪽, 1: 휠, 2:오른쪽 버튼 입니다

        Returns:
            bool: 마우스 버튼을 누르고 땟을때 True
        """
        if cls.mouse_click[button] == 1:
            return True
        return False